### modified pipeline script for the SPIRE Large Map observing mode.
### author: Pasquale Panuzzo, CEA Saclay, France, pasquale.panuzzo@cea.fr
### modified by Helene Roussel to be included in the Scanamorphos package
### This script is observation-dependent, to be adapted by the user.

###  This file has been adapted from the POF5_pipeline.py script
###  distributed as part of the Herschel Common Science System (HCSS).
###  Copyright 2001-2010 Herschel Science Ground Segment Consortium

###  HCSS is free software: you can redistribute it and/or modify
###  it under the terms of the GNU Lesser General Public License as
###  published by the Free Software Foundation, either version 3 of
###  the License, or (at your option) any later version.


#######################################################


## observation:
source = "atlas"
obsid = 0x500026A3L
nrep = 0
first_scan_num = 2

## For an observation consisting of a single scan, use nrep = 0 .
## Otherwise, use nrep = number of cross-scans (total number of scans = 2 * nrep).

## first_scan_num: The first scan of the observation will be given this index
## when saved to disk. Use first_scan_num = 1 in the general case.
## For parallel-mode observations where two scans are contained in different obsids,
## use first_scan_num = 2 for the second obsid.


## local settings:
poolName = "atlas_od192_500026A3"
path = "/data/spire/"
path_processed = path+source+"/processed_subscans/"

## Create the directory contained in the path_processed variable before saving the data.

#######################################################


## Import classes and definitions:
from herschel.spire.all import *
from herschel.spire.util import *
from herschel.ia.all import *
from herschel.ia.task.mode import *
from herschel.ia.pg import ProductSink
from java.lang import *
from java.util import *
from herschel.ia.obs.util import ObsParameter
from herschel.ia.pal.pool.lstore.util import TemporalPool
from herschel.spire.ia.pipeline.scripts.POF5.POF5_tasks import *
from herschel.spire.ia.pipeline.scripts.POF5.POF5_input import *

fits = FitsArchive()


## Read observation from pool: 
storage = ProductStorage(poolName)
result = storage.select(Query(ObservationContext))
print result
obs = result[0].product


## substitute most recent calibration context in the observation context:
cal = SpireCal.getInstance()
obs.calibration.update(cal)


## Extract the calibration products from the observation context:
bsmPos=obs.calibration.phot.bsmPos
lpfPar=obs.calibration.phot.lpfPar
detAngOff=obs.calibration.phot.detAngOff
elecCross=obs.calibration.phot.elecCross
optCross=obs.calibration.phot.optCross
chanTimeConst=obs.calibration.phot.chanTimeConst
chanNum=obs.calibration.phot.chanNum
fluxConvList=obs.calibration.phot.fluxConvList
tempDriftCorrList=obs.calibration.phot.tempDriftCorrList

## Extract the auxiliary products from the observation context:
hpp=obs.auxiliary.pointing
siam=obs.auxiliary.siam
timeCorr = obs.auxiliary.timeCorrelation


#######################################################


print "processing OBSID=", obs.obsid,"("+hex(obs.obsid)+")"

# start from level0:
level0_5=engConversion(obs.level0,cal=obs.calibration,tempStorage=Boolean.FALSE)

# start from level0.5 instead:
###level0_5=obs.level0.5


### to be modified when the number of scan legs is not the same for all scans
### within the obsid:

bbids = level0_5.getBbids(0xa103)
nsubscans = len(bbids)
if nrep == 0:
        nscans = 1
	nlegsperscan = nsubscans
else:
        nscans = 2*nrep
	nlegsperscan = nsubscans/nscans
print "number of scans: ", nscans
print "number of subscans: ", nsubscans
print "number of legs per scan = ", nlegsperscan


## Create Level1 context:
level1_allscans = Level1Context(obsid)

## Loop over scan legs:
counter = 0
for bbid in bbids:
#
	block=level0_5.get(bbid)
	print "processing BBID="+hex(bbid)
	## engineering data products:
	pdt  = block.pdt
	nhkt = block.nhkt
	if pdt == None:
		logger.severe("Building block "+hex(bbid)+" doesn't contain a PDT. Cannot process this building block.")
		print "Building block "+hex(bbid)+" doesn't contain a PDT. Cannot process this building block."
		continue
	if nhkt == None:
		logger.severe("Building block "+hex(bbid)+" doesn't contain a NHKT. Cannot process this building block.")
		print "Building block "+hex(bbid)+" doesn't contain a NHKT. Cannot process this building block."
		continue
#
	## attach turnaround data:
	bbCount=bbid & 0xFFFF
	pdtLead=None
	nhktLead=None
	pdtTrail=None
	nhktTrail=None
	if bbCount > 1:
		blockLead=level0_5.get(0xaf000000L+bbCount-1)
		pdtLead=blockLead.pdt
		nhktLead=blockLead.nhkt
		if pdtLead != None and pdtLead.sampleTime[-1] < pdt.sampleTime[0]-3.0:
			pdtLead=None
			nhktLead=None
	if bbid < MAX(Long1d(bbids)):
		blockTrail=level0_5.get(0xaf000000L+bbCount)
		pdtTrail=blockTrail.pdt
		nhktTrail=blockTrail.nhkt
		if pdtTrail != None and pdtTrail.sampleTime[0] > pdt.sampleTime[-1]+3.0:
			pdtTrail=None
			nhktTrail=None
	pdt=joinPhotDetTimelines(pdt,pdtLead,pdtTrail)
	nhkt=joinNhkTimelines(nhkt,nhktLead,nhktTrail)
#
	## BSM angles:
	bat=calcBsmAngles(nhkt,bsmPos=bsmPos)
#
	## pointing product:
	spp=createSpirePointing(detAngOff=detAngOff,bat=bat,hpp=hpp,siam=siam)
#
	## electrical crosstalk correction:
	pdt=elecCrossCorrection(pdt,elecCross=elecCross)
#
	## deglitching:
	pdt=waveletDeglitcher(pdt, scaleMin=1.0, scaleMax=8.0, scaleInterval=5, holderMin=-1.9, holderMax=-0.3, correlationThreshold=0.69)
#
	## electrical filter response correction:
	pdt=lpfResponseCorrection(pdt,lpfPar=lpfPar)
#
	## flux conversion:
	fluxConv=fluxConvList.getProduct(pdt.meta["biasMode"].value,pdt.startDate)
	pdt=photFluxConversion(pdt,fluxConv=fluxConv)
#
	## temperature drift correction:
###	tempDriftCorr=tempDriftCorrList.getProduct(pdt.meta["biasMode"].value,pdt.startDate)
###	pdt=temperatureDriftCorrection(pdt,tempDriftCorr=tempDriftCorr)
#
	## bolometer time response correction:
	pdt=bolometerResponseCorrection(pdt,chanTimeConst=chanTimeConst)
#
	## optical crosstalk correction:
	pdt=photOptCrossCorrection(pdt,optCross=optCross)
#
	## add pointing:
	psp=associateSkyPosition(pdt,spp=spp)
#
	## remove duplicate turnaround data:
	psp=cutPhotDetTimelines(psp,extend=True)
#
        ## apply the time correlation (???)
	psp=timeCorrelation(psp,timeCorr)
#
	## store product in level 1 product storage:
        level1_allscans.addProduct(psp)
#
	print "Completed BBID="+hex(bbid)
	counter = counter +1
	nsubsinit = nsubscans - counter
	if nsubsinit > 0:
 	  print str(nsubsinit) + " subscan(s) to process"


#######################################################


## Save level 1 data as fits files:

if nlegsperscan > 99:
	prefix_1="00"
        prefix_2="0"
else:
	prefix_1="0"
        prefix_2=""

for iscan in range(nscans):
        scan = first_scan_num + iscan
        if scan < 10 and nscans > 9:
		str_scan = "0"+str(scan)
        else:
		str_scan = str(scan)
	root_filename = path_processed+"level1_scan"+str_scan
#
	for ileg in range(nlegsperscan):
		bbid = ileg+iscan*nlegsperscan
		product = level1_allscans.getProduct(bbid)
                subs = bbid+1
                if subs < 10:
			str_subs = prefix_1+str(subs)
		else:
			if subs < 100:
				str_subs = prefix_2+str(subs)
			else:
				str_subs = str(subs)
		filename = root_filename+"_leg"+str_subs+".fits"
		fits.save(filename,product)
		print filename

### END OF SCRIPT

#######################################################
