### HIPE script to save PACS level-1 data to fits files for input to Scanamorphos
### supplied by Marc Sauvage

### note: Advanced scripts to process PACS data from level 0 to level 1 can be obtained
### from Marc Sauvage.


def convertL1ToScanam(frames,obsid=0l,cancelGlitch=0,assignRaDec=0):
    """
    This code converts level1 frames obtained within the context of the classical hipe 
    pipeline into level1 frames that can be saved to fits and ingested in Scanamorphos.
    
    convertL1ToScanam(frames,obsid=obsid)
    
    obsid is an optional parameter that you can use to provide the obsid if it is not
    in the metadata.
    
    The frames object is saved in the current directory under the name:
        [obsid]_[filter]_Scanam_level1Frames.fits
    """
    # HISTORY
    # 29-Jul-2010 MS Creation
    # 19-Jul-2011 MS Incorporate some options suggested by L. Calzoletti
    # Import section
    from herschel.ia.all import *
    from java.util import Date
    
    
    # Scanamorphos needs the astrometric information for all pixels, not only
    # for all frames. Check if the Ra and Dec cubes are present.
    if ((frames.containsKey('Ra') == False) and (assignRaDec == False)):
        print "ERROR -- The Ra and Dec cubes (coordinates for all pixels) are absent."
        print "ERROR -- Scanamorphos cannot process the data without this information."
        print "INFO  -- If you have access to the HSA, rerun this script with the option"
        print "INFO  -- assignRaDec=True.)"
        return
    elif (assignRaDec == True):
        print "INFO -- Connecting to HSA to obtain the observation context and thus"
        print "INFO -- the calibration tree that is relevant to this observation in order"
        print "INFO -- to compute the pixel coordinates correctly."
        obs = getObservation(obsid)
        calTree = getCalTree(obs)
        frames = photAssignRaDec(frames,calTree=calTree)
    
    
    # Scanamorphos needs three masks to be present in the frame structure.
    # It is intelligent enough for the masks to be in any order, but they
    # have to have the following names:
    # SA_badpix  for all bad pixels
    # NonScience for all the preamble data that occur before the scan
    # UNCLEANCHOP that contains a mask for the drifts occurring after calibration blocks
    print "Creating the Scanamorphos bad pixel mask and converting mask to boolean arrays", Date()
    # - badpixels is the or of BADPIXELS, BLINDPIXELS, SATURATION, MMT_Glitchmask, SecondGlitchmask
    
    if (cancelGlitch == False):
        origList = ['BADPIXELS','BLINDPIXELS','SATURATION','MMT_Glitchmask','SecondGlitchmask','2nd level glitchmask']
    else:
        print "INFO -- You have chosen to ignore the glitch information created in HIPE."
        print "INFO -- This is not a serious problem as Scanamorphos performs its own deglitching."
        origList = ['BADPIXELS','BLINDPIXELS','SATURATION']
    
    nMasks = len(origList)
    maskList = String1d()
    for i in range(nMasks):
        if (frames.mask.containsMask(origList[i])):
            maskList.append(origList[i])
    # now merge these masks and place them in the SA_badpix mask
    frames.mask.merge(maskList,'SA_badpix')
    # now check to see if the other two masks are present.
    #if (frames.mask.containsMask('NonScience') == False):
    #    print "WARNING -- The NonScience mask is absent"
    #    print "WARNING -- It flags preamble data occuring before the first leg"
    #    print "WARNING -- Scanamorphos will not crash but artifacts may be present"
    #if (frames.mask.containsMask('UNCLEANCHOP') == False):
    #    print "WARNING -- The UNCLEANCHOP mask is absent"
    #    print "WARNING -- (not necessary if you do not intend to mask transients)"
    
    print "Converting masks to boolean array that scanamorphos can read"
    frames.mask.convert2boolean()
    
    print "Saving level 1 frames converted for scanamorphos", Date()
    # I need some meta-data information to make up a name so still modifications to be done here.
    # check to see if the user has not specified the obsid
    if (obsid == 0l):
        if (frames.meta.containsKey('obsid')):
            obsid = frames.meta['obsid'].value
        else:
            print "WARNING -- obsid not found in the frames metadata - maybe they are old"
            obsid=0l
    
    if (frames.meta.containsKey('camName') == False):
        print "ERROR -- The key camName is not in the metadata."
        print "ERROR -- As this tells which side (blue/red) is present in the frames,"
        print "ERROR -- this is a really worrying sign regarding the data."
        print "ERROR -- Please investigate before trying Scanamorphos."
        return
    
    if (frames.meta['camName'].value == 'Red Photometer'):
        side = 'red'
    else:
        if (frames.meta['blue'].value == 'blue1'):
            side = 'blue'
        else:
            side = 'green'
    FitsArchive().save(str(obsid)+'_'+side+'_Scanam_level1Frames.fits',frames)
