pro convert_hcssfits_pacs, dir_in=dir_in, dir_out=dir_out, root=root, color=color, $
                           onescan=onescan, nsubs=nsubs, mask_transient=mask_transient, $
                           advanced_masks=advanced_masks


;;; conversion of Hipe level-1 fits files into Scanamorphos input structures, for PACS

;;; to be used one observation at a time: root name must include the obsid number
;;; convention Xra = -RA, Ydec = DEC

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; dir_in: directory where the Hipe file(s) can be found
;;; dir_out (optional): directory where the Scanamorphos input structures will be saved
;;;                     (= dir_in by default)
;;; root: root name of the Hipe file(s)
;;;       must allow to uniquely select the file associated with the obsid to process
;;;       (or the files if the observation was split into several parts)
;;; color: either "blue", "green", or "red"
;;;        (i.e. 70, 100, or 160 microns)
;;;        Please do not confuse blue and green observations
;;;        (same array but different filters and different parameters).
;;; onescan (optional): Normally, the "repetition" metadata keyword should be incremented
;;;                     only when a new scan has started. If it is wrongly incremented
;;;                     within a scan, select this option to specify that the obsid contains
;;;                     only one scan.
;;; nsubs (optional): Normally, one Hipe file contains a whole scan or several scans.
;;;                   If a scan is split between several Hipe files, set the "nsubs" parameter
;;;                   to the number of these files, and give them the same root name.
;;; mask_transient (optional): Select this option if you want to mask transients after
;;;                            calibration flashes in SDP or pre-nominal data
;;;                            (now possible even if the advanced masks are undefined;
;;;                             the mask duration is between a fourth and half a scan leg).
;;; advanced_masks: kept for backward compatibility purposes.
;;;                 Masks created by convertL1ToScanam.py are now automatically propagated.


;;; Before running Scanamorphos, include the full path name of dir_out
;;; in the first line of SCANAMORPHOS/scanlist_pacs
;;; and update the list of input structures, one per line.

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel and Marc Sauvage
;;; H.R.: roussel@iap.fr
;;;       Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France
;;; M.S.: marc.sauvage@cea.fr
;;;       CEA, Laboratoire AIM, Irfu/SAp, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; HISTORY
;;; MS: I try some modifications to also get the mask information
;;; MS: June 2 - 2006 The mask is modified in hipe so now we get 3 here



resolve_all, /continue_on_error, /quiet
crs = string(13b) + string(10b)

n_pcaltr = 200

nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
if (not keyword_set(dir_out)) then dir_out = dir_in
nc = strlen(dir_out)
if strmid(dir_out, nc-1) ne '/' then dir_out = dir_out + '/'


if (not keyword_set(color)) then color = ''
if (color eq 'blue' or color eq 'green') then nb = 64L*32L $
   else if color eq 'red' then nb = 32L*16L else begin
      print, 'no array selected: exit'
      print, 'should be "blue", "green" or "red"'
      goto, end_label
   endelse

if (not keyword_set(nsubs)) then nsubs = 1


; MS the signal and status are indeed the first two extensions
; so they can be defined here.
exten_signal = 1
exten_status = 2   ;;; for time vector

; Ra and Dec will appear at different positions because the
; mask dataset is of variable size (more than one cube). 
; So it is better to look for their positions later.



name_search = dir_in + root + '*.fits'
files = file_search(name_search, /fold_case, /test_read, count=nf)
if nf ne nsubs then begin
   if nsubs eq 1 then print, crs + 'number of observation(s) found is not 1: exit' $
      else print, crs + 'number of scan parts found is not ' + chain(nsubs) + ': exit'
   print, 'Check directory and file names and permissions.'
   goto, end_label
endif else begin
   print, crs + 'observation found !' + crs
   for k = 0, nsubs - 1 do begin
      file = files(k)

      ; MS modifications
      ; The header lists the number of dataset stored in the
      ; extensions but one dataset can require more than one
      ; extension to be stored. In the header we have the
      ; relation between dataset number and extension number
      ; the masks are always one data set so Ra and Dec are the 13th
      ; and 14th dataset.  
      ; read the header of extension 0 (main header)
      main_hdr = headfits(file, exten=0)

      ; MS 02/08/10 - I now make it more robust: get all the extension names
      ; and find the location of those I'm interested in.

      ; this is the number of datasets
      ndataset = sxpar(main_hdr, 'dsets___')
      ; now get all their names.
      allnames_tab = strarr(ndataset)
      for iset = 0, ndataset - 1 do begin
         this_exten = sxpar(main_hdr, 'DS_'+chain(iset))
         allnames_tab(iset) = strtrim(sxpar(headfits(file, exten=this_exten), 'EXTNAME'), 2)
      endfor

      ; The datasets I need at this point are called 'Ra', 'Dec', and 'BooleanMasks'
      index = where(allnames_tab eq 'Ra', count)
      if (count ne 1) then begin
         print, 'ERROR -- Ra extension not found. This is fatal!'
         has_ra = 0
      endif else begin
         has_ra = 1
         ra_keyname = 'DS_' + chain(index(0))
         exten_ra = sxpar(main_hdr, ra_keyname)
      endelse
      index = where(allnames_tab eq 'Dec', count)
      if (count ne 1) then begin
         print, 'ERROR -- Dec extension not found. This is fatal!'
         has_dec = 0
      endif else begin
         has_dec = 1
         dec_keyname = 'DS_' + chain(index(0))
         exten_dec = sxpar(main_hdr, dec_keyname)
      endelse
      if (not(has_ra) or not(has_dec)) then goto, end_label
      index = where(allnames_tab eq 'BooleanMasks', count)
      if (count ne 1) then begin
         has_masks = 0
         print, 'ERROR - The boolean versions of the masks are not present.'
         print, 'ERROR - No mask information will be propagated to Scanamorphos.'
      endif else begin
         has_masks = 1
         ; use of the prefix 'last' is because this is usually the last one
         last_keyname = 'DS_' + chain(index(0))
         last_ext = sxpar(main_hdr, last_keyname)
      endelse

      ;;; use of advanced masks now by default:
         ; Now let's try to find the extension number of the 
         ; mask needed for scanamorphos (created in hipe)
         ; We basically use 3: SA_badpix is all the bad pixels
         ;                                  glitches, saturation...
         ;                     NonScience flags the preamble of the obs
         ;                     UNCLEANCHOP is the cal transient
         ;                                 not the best name.
         ; now last_ext is the extension number for the masks dataset
         ; so I get the header of that last exention
         mask_hdr = headfits(file, exten=last_ext)
         ; now find the number of datasets saved for the masks
         nsubdataset = sxpar(mask_hdr, 'dsets___')
         ; get the names of all the masks
         names_tab = strarr(nsubdataset)
         for iset = 0, nsubdataset - 1 do begin
            this_exten = sxpar(mask_hdr,'DS_'+chain(iset))
            names_tab(iset) = strtrim(sxpar(headfits(file, exten=this_exten),'EXTNAME'), 2)
         endfor

         ; bad pixels mask, called SA_badpix
         index = where(names_tab eq 'SA_badpix', count)
         if (count ne 1) then begin
            print, 'ERROR - The bad pixels mask is not found.'
            print, '   Either run convertL1ToScanam.py before calling convert_hcssfits_pacs,'
            print, '   or set bad pixels to NaN within HIPE.'
            has_badpix = 0
         endif else begin
            has_badpix = 1
            this_exten = sxpar(mask_hdr, 'DS_'+chain(index(0)))
            maskBadPixel = read_hcss(file, this_exten)
         endelse

         ; NonScience mask, masking the preamble to the actual scan
         index = where(names_tab eq 'NonScience', count)
         if (count ne 1) then has_nonscience = 0 else begin
            has_nonscience = 1
            this_exten = sxpar(mask_hdr, 'DS_'+chain(index(0)))
            maskNonScience = read_hcss(file, this_exten)
         endelse
         ;
         ; end of MS modifications
         ; The mask information is dealt with after reading the signal.
      ;;;

 
      scan_status = read_hcss(file, exten_status)
      scan_status = scan_status.__DATA__
      if k eq 0 then time_start = min(scan_status.finetime)
      t = 1.d-6 * double(scan_status.finetime - time_start)
      nt = n_elements(t)
      if k eq 0 then time = t else time = [time, t]

      tags_status = strlowcase(tag_names(scan_status))
      w = where(tags_status eq 'repetition', cw)
      if (cw eq 1 and (not keyword_set(onescan))) then begin
         ind = scan_status.repetition
         ind -= min(ind)
         nscans = max(ind) + 1
      endif else begin
         nscans = 1
         ind = intarr(nt)
      endelse
      if k eq 0 then ind_scan = ind else ind_scan = [ind_scan, ind]


      bbt = scan_status.bbtype
      w = where(bbt gt 0 and bbt lt 3282, cw)
      if cw ge 1 then print, 'excluding ' + chain(cw) + ' time step(s) on a total of ' + chain(nt) $
         + ': non-nominal frames'

      ;;; initial slew:
      wscan = where(bbt gt 3281)
      if wscan(0) gt 0 then begin
         w = where(bbt(0:wscan(0)-1) lt 1, cw)
         if cw ge 1 then begin
            print, 'excluding ' + chain(cw) + ' time step(s): initial slew'
            bbt(w) = 2
         endif
      endif

      ;;; slews between scans:
      if nt / nscans gt 1500 then begin
         for s = 0, nscans - 1 do begin
            ws = where(ind eq s, cs)
            flag_nom = intarr(cs)
            w = where(bbt(ws) gt 3281, cw)
            if cw ge 10 then begin
               flag_nom(w) = 1
               w = where(flag_nom(0:cs-2) eq 1 and flag_nom(1:cs-1) eq 0)
               we = w(0)
               w = where(flag_nom(0:we) eq 0, cw)
               if cw ge 1 then wi = w(cw-1) + 1 else wi = 0
               if we - wi + 1 lt 10 then bbt(ws(wi:we)) = 0
               w = where(flag_nom(0:cs-2) eq 0 and flag_nom(1:cs-1) eq 1, cw)
               wi = w(cw-1) + 1
               w = where(flag_nom(wi:cs-1) eq 0, cw)
               if cw ge 1 then we = wi + w(0) - 1 else we = cs - 1
               if we - wi + 1 lt 10 then bbt(ws(wi:we)) = 0
            endif
         endfor
         nslew = 0L
         for s = 1, nscans - 1 do begin
            ws = where(ind eq s, cs)
            wnom = where(bbt(ws) eq 3282, cw)
            if cw ge 10 then begin
               w = where(bbt(ws) eq 0)
               if wnom(0) gt w(0) then begin
                  wns = where(bbt(ws(w(0):cs-1)) gt 0)
                  bbt(ws(w(0):w(0)+wns(0)-1)) = 2
                  nslew += wns(0)
               endif
               ws = where(ind eq s - 1, cs)
               w = where(bbt(ws) eq 0, cw)
               wnom = where(bbt(ws) eq 3282, cnom)
               if wnom(cnom-1) lt w(cw-1) then begin
                  wns = where(bbt(ws(0:w(cw-1))) gt 0, cns)
                  bbt(ws(wns(cns-1)+1:w(cw-1))) = 2
                  nslew += (w(cw-1) - wns(cns-1))
               endif
            endif
         endfor
         if nslew gt 0 then print, 'excluding ' + chain(nslew) + ' time step(s): slew(s) between scans'
      endif

      ;;; frames sandwiched between PCAL flashes and subsequent non-nominal data, PCAL transients:
      wcal = where(bbt eq 4 or bbt eq 1103, ccal)
      while ccal ge 1 do begin
         wnom = where(bbt(wcal(0):*) eq 0 or bbt(wcal(0):*) eq 3282 or bbt(wcal(0):*) eq 2)
         wis = where(bbt(wcal(0):*) eq 3282, ci)
         if ci lt 1 then bbt(wcal(0):*) = 3 else begin
            wes = where(bbt(wcal(0)+wis(0):*) lt 3282, ce)
            if ce ge 1 then dn_s = wes(0) else dn_s = nt - wcal(0) - wis(0)
            dn_min = min([100, dn_s / 10])
            wz = where(bbt(wcal(0)+wnom(0):*) gt 0 and bbt(wcal(0)+wnom(0):*) lt 3282 and bbt(wcal(0)+wnom(0):*) ne 2, cz)
            wstop = wcal(0) + wnom(0) - 1
            if cz ge 1 then if wz(0) lt dn_min then begin
               print, 'excluding ' + chain(wz(0)) + ' time step(s): after PCAL flash and before nominal data'
               wstop = wcal(0) + wnom(0) + wz(0) - 1
               wze = where(bbt(wcal(0)+wnom(0)+wz(0):*) eq 0 or bbt(wcal(0)+wnom(0)+wz(0):*) eq 3282 $
                 or bbt(wcal(0)+wnom(0)+wz(0):*) eq 2, cze)
               if cze ge 1 then wstop += wze(0)
            endif
            bbt(wcal(0):wstop) = 3
            ;;; transients after calibration flash:
            if keyword_set(mask_transient) then begin
               n_mask = min([max([n_pcaltr, dn_s / 4]), dn_s / 2])
               wm_start = where(bbt(wstop+1:*) eq 3282, cwm)
               if cwm lt 1 then wm_start = 0 else wm_start = wm_start(0)
               wm = where(bbt(wstop+1:wstop+1+wm_start+n_mask) eq 0 $
                  or bbt(wstop+1:wstop+1+wm_start+n_mask) eq 3282, cwm)
               if cwm ge 1 then bbt(wstop+1+wm) = -1
            endif
         endelse
         wcal = where(bbt eq 4 or bbt eq 1103, ccal)
      endwhile

      if k eq 0 then bbtype = bbt else bbtype = [bbtype, bbt]


      band = scan_status.band
      ; sometimes the first values are different because 
      ; we set the filter during the observation.
      ; using the middle point is better.
      print, ''
      filter = band(n_elements(band) / 2)
      print, 'wavelength band: ' + filter
      if ((strupcase(strmid(filter, 0, 1)) eq 'B' and color eq 'red') $
         or (strupcase(strmid(filter, 0, 1)) eq 'R' and color ne 'red')) then begin
         print, 'selected color does not match the array: exit'
         goto, end_label
      endif
      if (strupcase(filter) eq 'BS' and color eq 'green') then begin
         print, 'Changing the color from green to blue to match the data.'
         color = 'blue'
      endif
      if (strupcase(filter) eq 'BL' and color eq 'blue') then begin
         print, 'Changing the color from blue to green to match the data.'
         color = 'green'
      endif
      print, ''


      scan_signal = read_hcss(file, exten_signal)
      scan_signal = scan_signal.__DATA__
      sig = reform(scan_signal, nt, nb)


      if has_masks then begin
         ;;; use mask information here before concatenating the signal:
                                ; MS Modif: I mask in the signal all
                                ; the bad pixels as well as the non
                                ; science pixels. By default, the
                                ; transient after the calibration
                                ; block is not masked.
         ; bad pixels
         if (has_badpix) then begin
            scan_mask = reform(maskBadPixel.__DATA__, nt, nb)
            wz = where(scan_mask eq 1, cz)
            if cz ge 1 then sig(wz) = !values.f_nan
         endif
         ; Non science data
         if (has_nonscience) then begin
            scan_mask = reform(maskNonScience.__DATA__, nt, nb)
            wz = where(scan_mask eq 1, cz)
            if cz ge 1 then sig(wz) = !values.f_nan
         endif
      endif

      if keyword_set(mask_transient) then begin
         wz = where(bbt lt 0, cz)
         if cz ge 1 then sig(wz, *) = !values.f_nan
      endif

      if k eq 0 then signal = sig else signal = [signal, sig]


      scan_ra_deg = read_hcss(file, exten_ra)
      scan_ra_deg = scan_ra_deg.__DATA__
      Xra = -reform(scan_ra_deg, nt, nb)
      scan_dec_deg = read_hcss(file, exten_dec)
      scan_dec_deg = scan_dec_deg.__DATA__
      Ydec = reform(scan_dec_deg, nt, nb)
      if k eq 0 then begin
         Xra_bolos = Xra
         Ydec_bolos = Ydec
      endif else begin
         Xra_bolos = [Xra_bolos, Xra]
         Ydec_bolos = [Ydec_bolos, Ydec]
      endelse
   endfor


   ;;; exclude time jumps:

   nt = n_elements(time)
   diff_t = [1., time(1:nt-1) - time(0:nt-2)]
   w1 = where(diff_t le 0., cw)
   if cw ge 1 then begin
      flag = intarr(nt)
      for i = 0, cw - 1 do begin
         if w1(i) ge 2 then off_back = 2 else off_back = 1
         w2 = w1(i) + where(time(w1(i):*) gt time(w1(i)-off_back), c2)
         while (w2(0) gt w1(i) + 30 and w1(i) ge off_back + 1) do begin
            off_back += 1
            w2 = w1(i) + where(time(w1(i):*) gt time(w1(i)-off_back), c2)
         endwhile
         if c2 ge 1 then flag(w1(i)-off_back+1:w2(0)-1) = 1 else flag(w1(i)-1:*) = 1
      endfor

      wkeep = where(flag lt 1, nt, ncomplement=cz)
      print, 'excluding ' + chain(cz) + ' time step(s) (absurd time vector)'
      time = time(wkeep)
      ind_scan = ind_scan(wkeep)
      bbtype = bbtype(wkeep)
      signal = signal(wkeep, *)
      Xra_bolos = Xra_bolos(wkeep, *)
      Ydec_bolos = Ydec_bolos(wkeep, *)
   endif


   ;;; mask invalid bolometers:

   w = where(finite(signal) eq 0 or finite(Xra_bolos) eq 0 or finite(Ydec_bolos) eq 0, cw)
   if cw ge 1 then begin
      wb = w / nt
      histo = histogram(wb, min=0., max=nb-1)
      wkeep = where(histo lt 9 * nt / 10, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' bolometers'
         signal(*, wz) = !values.f_nan
         print, ''
      endif
   endif



   ;;; loop on scans:

   for s = 0, nscans - 1 do begin

      ;;; to select scan s and exclude spurious calibration data:
      ws = where(ind_scan eq s and (bbtype lt 1 or bbtype gt 3281), cs)

      if cs ge 100 then begin
         print, 'scan ' + chain(s) + ': ' + chain(cs) + ' time samples'
         time_s = time(ws)

         ;;; create structure:
         if color eq 'blue' then begin
            signal_psw = signal(ws, *)
            Xra_bolos_psw = Xra_bolos(ws, *)
            Ydec_bolos_psw = Ydec_bolos(ws, *)
            scan = create_struct('time_s', time_s, 'signal_psw', signal_psw, $
               'Xra_bolos_psw', Xra_bolos_psw, 'Ydec_bolos_psw', Ydec_bolos_psw, $
               'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
         endif else if color eq 'green' then begin
            signal_pmw = signal(ws, *)
            Xra_bolos_pmw = Xra_bolos(ws, *)
            Ydec_bolos_pmw = Ydec_bolos(ws, *)
            scan = create_struct('time_s', time_s, 'signal_pmw', signal_pmw, $
               'Xra_bolos_pmw', Xra_bolos_pmw, 'Ydec_bolos_pmw', Ydec_bolos_pmw, $
               'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
         endif else if color eq 'red' then begin
            signal_plw = signal(ws, *)
            Xra_bolos_plw = Xra_bolos(ws, *)
            Ydec_bolos_plw = Ydec_bolos(ws, *)
            scan = create_struct('time_s', time_s, 'signal_plw', signal_plw, $
               'Xra_bolos_plw', Xra_bolos_plw, 'Ydec_bolos_plw', Ydec_bolos_plw, $
               'units', ['times in seconds', 'fluxes in Jy/array pixel', 'coordinates in degrees'])
         endif

         file_out = 'stru_' + root + '_scan' + chain(s+1) + '.xdr'
         save, filename=dir_out+file_out, scan, /xdr
         print, 'saved file "' + file_out + '"'
      endif
   endfor
   print, ''

endelse


end_label:

end
