pro convert_hcssfits_spire, dir_in=dir_in, dir_out=dir_out, root=root, $
                            offset_ra_as=offset_ra_as, offset_dec_as=offset_dec_as


;;; conversion of Hipe level-1 fits files into Scanamorphos input structures, for SPIRE

;;; to be used one scan at a time: root name must include the scan number
;;; convention Xra = -RA, Ydec = DEC

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; dir_in: directory where the Hipe files can be found
;;; dir_out (optional): directory where the Scanamorphos input structures will be saved
;;;                     (= dir_in by default)
;;; root: root name of the Hipe files
;;;       must allow to uniquely select all the files associated with the scan to process
;;;       (i.e. all the legs of the scan)
;;; offset_ra_as and offset_dec_as (optional):
;;;      If offsets between a reference observation and the input observation have been
;;;      measured, they can be supplied by these parameters to correct the astrometry.
;;;      They will be added to the input coordinates
;;;      (in arcsec, counted positively toward East and toward North)


;;; Before running Scanamorphos, include the full path name of dir_out
;;; in the first line of SCANAMORPHOS/scanlist_spire
;;; and update the list of input structures, one per line.

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



resolve_all, /continue_on_error, /quiet

nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
if (not keyword_set(dir_out)) then dir_out = dir_in
nc = strlen(dir_out)
if strmid(dir_out, nc-1) ne '/' then dir_out = dir_out + '/'


npsw = 139
npmw = 88
nplw = 43


exten_signal = 3
exten_ra = 4
exten_dec = 5
exten_mask = 1



name_search = dir_in + root + '*.fits'
files = file_search(name_search, /fold_case, /test_read, count=nf)
if nf ge 1 then begin
   print, chain(nf) + ' subscans found'


   ;;; check if subscans are time-ordered:

   tstart = fltarr(nf)
   tstart_off = fltarr(nf)
   for i = 0, nf - 1 do begin
      subscan_signal = read_hcss(files(i), exten_signal)
      subscan_signal = subscan_signal.__DATA__
      tstart(i) = subscan_signal(0).sampletime
      tstart_off(i) = min(subscan_signal.sampletime - tstart(0))
   endfor
   ind = sort(tstart_off)
   files = files(ind)
   print, 'starting time: ' + chain(tstart(ind(0)))


   ;;; read data:

   for i = 0, nf - 1 do begin
      subscan_signal = read_hcss(files(i), exten_signal)
      subscan_signal = subscan_signal.__DATA__
      tags = tag_names(subscan_signal)
      tags_short = strmid(tags, 0, 3)

      if i eq 0 then time_s = subscan_signal.sampletime $
         else time_s = [time_s, subscan_signal.sampletime]
      nt = n_elements(subscan_signal.sampletime)

      w_psw = where(tags_short eq 'PSW' and tags ne 'PSWR1' and tags ne 'PSWT1' and tags ne 'PSWT2' $
         and tags ne 'PSWDP1' and tags ne 'PSWDP2', c_psw)
      if c_psw ne npsw then begin
         print, 'npsw not equal to ' + chaine(npsw) + ' !'
         stop
      endif
      tags_psw = tags(w_psw)
      signal = fltarr(nt, npsw)
      for b = long(0), npsw - 1 do signal(*, b) = subscan_signal.(w_psw(b))
      if i eq 0 then signal_psw = signal else signal_psw = [signal_psw, signal]

      w_pmw = where(tags_short eq 'PMW' and tags ne 'PMWR1' and tags ne 'PMWT1' and tags ne 'PMWT2' $
         and tags ne 'PMWDP1' and tags ne 'PMWDP2', c_pmw)
      if c_pmw ne npmw then begin
         print, 'npmw not equal to ' + chaine(npmw) + ' !'
         stop
      endif
      tags_pmw = tags(w_pmw)
      signal = fltarr(nt, npmw)
      for b = long(0), npmw - 1 do signal(*, b) = subscan_signal.(w_pmw(b))
      if i eq 0 then signal_pmw = signal else signal_pmw = [signal_pmw, signal]

      w_plw = where(tags_short eq 'PLW' and tags ne 'PLWR1' and tags ne 'PLWT1' and tags ne 'PLWT2' $
         and tags ne 'PLWDP1' and tags ne 'PLWDP2', c_plw)
      if c_plw ne nplw then begin
         print, 'nplw not equal to ' + chaine(nplw) + ' !'
         stop
      endif
      tags_plw = tags(w_plw)
      signal = fltarr(nt, nplw)
      for b = long(0), nplw - 1 do signal(*, b) = subscan_signal.(w_plw(b))
      if i eq 0 then signal_plw = signal else signal_plw = [signal_plw, signal]


      subscan_ra_deg = read_hcss(files(i), exten_ra)
      subscan_ra_deg = subscan_ra_deg.__DATA__
      subscan_dec_deg = read_hcss(files(i), exten_dec)
      subscan_dec_deg = subscan_dec_deg.__DATA__
      tags_ra = tag_names(subscan_ra_deg)
      tags_dec = tag_names(subscan_dec_deg)

      Xra_bolos = dblarr(nt, npsw)
      Ydec_bolos = dblarr(nt, npsw)
      for b = long(0), npsw - 1 do begin
         w = where(tags_ra eq tags_psw(b))
         Xra_bolos(*, b) = -subscan_ra_deg.(w)
         w = where(tags_dec eq tags_psw(b))
         Ydec_bolos(*, b) = subscan_dec_deg.(w)
      endfor
      if i eq 0 then begin
         Xra_bolos_psw = Xra_bolos
         Ydec_bolos_psw = Ydec_bolos
      endif else begin
         Xra_bolos_psw = [Xra_bolos_psw, Xra_bolos]
         Ydec_bolos_psw = [Ydec_bolos_psw, Ydec_bolos]
      endelse

      Xra_bolos = dblarr(nt, npmw)
      Ydec_bolos = dblarr(nt, npmw)
      for b = long(0), npmw - 1 do begin
         w = where(tags_ra eq tags_pmw(b))
         Xra_bolos(*, b) = -subscan_ra_deg.(w)
         w = where(tags_dec eq tags_pmw(b))
         Ydec_bolos(*, b) = subscan_dec_deg.(w)
      endfor
      if i eq 0 then begin
         Xra_bolos_pmw = Xra_bolos
         Ydec_bolos_pmw = Ydec_bolos
      endif else begin
         Xra_bolos_pmw = [Xra_bolos_pmw, Xra_bolos]
         Ydec_bolos_pmw = [Ydec_bolos_pmw, Ydec_bolos]
      endelse

      Xra_bolos = dblarr(nt, nplw)
      Ydec_bolos = dblarr(nt, nplw)
      for b = long(0), nplw - 1 do begin
         w = where(tags_ra eq tags_plw(b))
         Xra_bolos(*, b) = -subscan_ra_deg.(w)
         w = where(tags_dec eq tags_plw(b))
         Ydec_bolos(*, b) = subscan_dec_deg.(w)
      endfor
      if i eq 0 then begin
         Xra_bolos_plw = Xra_bolos
         Ydec_bolos_plw = Ydec_bolos
      endif else begin
         Xra_bolos_plw = [Xra_bolos_plw, Xra_bolos]
         Ydec_bolos_plw = [Ydec_bolos_plw, Ydec_bolos]
      endelse


      subscan_mask = read_hcss(files(i), exten_mask)
      subscan_mask = subscan_mask.__DATA__
      tags = tag_names(subscan_mask)
      mask = lonarr(nt, npsw)
      for b = long(0), npsw - 1 do begin
         w = where(tags eq tags_psw(b))
         mask(*, b) = subscan_mask.(w)
      endfor
      if i eq 0 then mask_psw = mask else mask_psw = [mask_psw, mask]

      mask = lonarr(nt, npmw)
      for b = long(0), npmw - 1 do begin
         w = where(tags eq tags_pmw(b))
         mask(*, b) = subscan_mask.(w)
      endfor
      if i eq 0 then mask_pmw = mask else mask_pmw = [mask_pmw, mask]

      mask = lonarr(nt, nplw)
      for b = long(0), nplw - 1 do begin
         w = where(tags eq tags_plw(b))
         mask(*, b) = subscan_mask.(w)
      endfor
      if i eq 0 then mask_plw = mask else mask_plw = [mask_plw, mask]
   endfor

   nt = n_elements(time_s)

   if keyword_set(offset_ra_as) then begin
      Xra_bolos_psw -= (offset_ra_as / 3600.D0)
      Xra_bolos_pmw -= (offset_ra_as / 3600.D0)
      Xra_bolos_plw -= (offset_ra_as / 3600.D0)
   endif
   if keyword_set(offset_dec_as) then begin
      Ydec_bolos_psw += (offset_dec_as / 3600.D0)
      Ydec_bolos_pmw += (offset_dec_as / 3600.D0)
      Ydec_bolos_plw += (offset_dec_as / 3600.D0)
   endif


   ;;; grossly estimate the noise of each bolometer (indicative only, not used to mask bolometers):
   dev = !values.f_nan + fltarr(npsw)
   for b = long(0), npsw - 1 do begin
      signal = signal_psw(*, b)
      w = where(finite(signal) gt 0, cw)
      if cw ge 1 then begin
         base = smooth(signal, 31, /edge_truncate)
         signal -= base
         med = median(signal(w))
         dev(b) = sqrt(mean((signal(w) - med)^2.))
         if dev(b) gt 1.e-6 then for k = 0, 9 do begin
            wdef = where(abs(signal(w) - med) lt 6. * dev(b), cdef)
            if cdef ge 1 then begin
               med = median(signal(w(wdef)))
               dev(b) = sqrt(mean((signal(w(wdef)) - med)^2.))
            endif
         endfor
      endif
   endfor
   med_dev = median(dev)
   dev_orig = dev
   w = where(finite(dev_orig) lt 1, cw)
   if cw ge 1 then dev_orig(w) = 0.



   ;;; eliminate slow bolometers:
   w = where(tags_psw eq 'PSWA10' or tags_psw eq 'PSWA11' or tags_psw eq 'PSWA13')
   dev(w) = -9.

   dev = dev ## (fltarr(nt) + 1.)


   ;;; create a global mask taking into account only fatal mask bits:
   set_mask_spire, mask_psw, mask_psw_fatal, nt=nt, nb=npsw


   w = where(finite(signal_psw) eq 0 or finite(Xra_bolos_psw) eq 0 or finite(Ydec_bolos_psw) eq 0 $
      or mask_psw_fatal gt 0 or dev lt 0., cw)

   if cw ge 1 then begin
      signal_psw(w) = !values.f_nan
      wb = w / nt
      histo = histogram(wb, min=0., max=npsw-1)
      wkeep = where(histo lt 9 * nt / 10, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' PSW bolometer(s):' + strjoin('   ' + tags_psw(wz))
         print, '                           ' + strjoin('   ' + chain(wz))
         signal_psw(*, wz) = !values.f_nan
         print, 'noise/med_noise = ' + strjoin('   ' + chain(dev_orig(wz) / med_dev, 2))
         print, ''
      endif
   endif



   dev = !values.f_nan + fltarr(npmw)
   for b = long(0), npmw - 1 do begin
      signal = signal_pmw(*, b)
      w = where(finite(signal) gt 0, cw)
      if cw ge 1 then begin
         base = smooth(signal, 31, /edge_truncate)
         signal -= base
         med = median(signal(w))
         dev(b) = sqrt(mean((signal(w) - med)^2.))
         if dev(b) gt 1.e-6 then for k = 0, 9 do begin
            wdef = where(abs(signal(w) - med) lt 6. * dev(b), cdef)
            if cdef ge 1 then begin
               med = median(signal(w(wdef)))
               dev(b) = sqrt(mean((signal(w(wdef)) - med)^2.))
            endif
         endfor
      endif
   endfor
   med_dev = median(dev)
   dev_orig = dev
   w = where(finite(dev_orig) lt 1, cw)
   if cw ge 1 then dev_orig(w) = 0.


   ;;; slow bolometers:
   w = where(tags_pmw eq 'PMWA13')
   dev(w) = -9.

   dev = dev ## (fltarr(nt) + 1.)


   ;;; create a global mask taking into account only fatal mask bits:
   set_mask_spire, mask_pmw, mask_pmw_fatal, nt=nt, nb=npmw


   w = where(finite(signal_pmw) eq 0 or finite(Xra_bolos_pmw) eq 0 or finite(Ydec_bolos_pmw) eq 0 $
      or mask_pmw_fatal gt 0 or dev lt 0., cw)

   if cw ge 1 then begin
      signal_pmw(w) = !values.f_nan
      wb = w / nt
      histo = histogram(wb, min=0., max=npmw-1)
      wkeep = where(histo lt 9 * nt / 10, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' PMW bolometer(s):' + strjoin('   ' + tags_pmw(wz))
         print, '                           ' + strjoin('   ' + chain(wz))
         signal_pmw(*, wz) = !values.f_nan
         print, 'noise/med_noise = ' + strjoin('   ' + chain(dev_orig(wz) / med_dev, 2))
         print, ''
      endif
   endif



   dev = !values.f_nan + fltarr(nplw)
   for b = long(0), nplw - 1 do begin
      signal = signal_plw(*, b)
      w = where(finite(signal) gt 0, cw)
      if cw ge 1 then begin
         base = smooth(signal, 31, /edge_truncate)
         signal -= base
         med = median(signal(w))
         dev(b) = sqrt(mean((signal(w) - med)^2.))
         if dev(b) gt 1.e-6 then for k = 0, 9 do begin
            wdef = where(abs(signal(w) - med) lt 6. * dev(b), cdef)
            if cdef ge 1 then begin
               med = median(signal(w(wdef)))
               dev(b) = sqrt(mean((signal(w(wdef)) - med)^2.))
            endif
         endfor
      endif
   endfor
   med_dev = median(dev)
   dev_orig = dev
   w = where(finite(dev_orig) lt 1, cw)
   if cw ge 1 then dev_orig(w) = 0.


   ;;; no slow bolometers

   dev = dev ## (fltarr(nt) + 1.)

   ;;; create a global mask taking into account only fatal mask bits:
   set_mask_spire, mask_plw, mask_plw_fatal, nt=nt, nb=nplw


   w = where(finite(signal_plw) eq 0 or finite(Xra_bolos_plw) eq 0 or finite(Ydec_bolos_plw) eq 0 $
      or mask_plw_fatal gt 0 or dev lt 0., cw)

   if cw ge 1 then begin
      signal_plw(w) = !values.f_nan
      wb = w / nt
      histo = histogram(wb, min=0., max=nplw-1)
      wkeep = where(histo lt 9 * nt / 10, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' PLW bolometer(s):' + strjoin('   ' + tags_plw(wz))
         print, '                           ' + strjoin('   ' + chain(wz))
         signal_plw(*, wz) = !values.f_nan
         print, 'noise/med_noise = ' + strjoin('   ' + chain(dev_orig(wz) / med_dev, 2))
         print, ''
      endif
   endif



   ;;; create structure:
   scan = create_struct('time_s', time_s, $
      'signal_psw', signal_psw, 'signal_pmw', signal_pmw, 'signal_plw', signal_plw, $
      'Xra_bolos_psw', Xra_bolos_psw, 'Ydec_bolos_psw', Ydec_bolos_psw, $
      'Xra_bolos_pmw', Xra_bolos_pmw, 'Ydec_bolos_pmw', Ydec_bolos_pmw, $
      'Xra_bolos_plw', Xra_bolos_plw, 'Ydec_bolos_plw', Ydec_bolos_plw, $
      'units', ['times in seconds', 'fluxes in Jy/beam', 'coordinates in degrees'])

   file_out = 'stru_' + root + '.xdr'
   save, filename=dir_out+file_out, scan, /xdr
   print, 'saved file "' + file_out + '"'
   print, ''

   w = where(tags_psw eq 'PSWE8')
   print, 'index of PSW center bolometer: ' + chain(w)

   w = where(tags_pmw eq 'PMWD6')
   print, 'index of PMW center bolometer: ' + chain(w)

   w = where(tags_plw eq 'PLWC5')
   print, 'index of PLW center bolometer: ' + chain(w)
   print, ''

endif else begin
   print, ''
   print, 'no subscan found'
   print, 'Check directory and file names and permissions.'
endelse


end
