;+
; NAME: READ_HCSS
;
; PURPOSE: reads an HCSS product stored as FITS file.
;
; CALLING SEQUENCE:
;  product=read_hcss(fitsfile, [extension])
;   
; INPUTS:
;  fitsfile -- STRING : file name of the FITS file
;  extension -- INTEGER : extension table index to be read
;
; OUTPUTS:
;  product -- structure : contains all info of the input product
;
; MODIFICATION HISTORY:
;   20-May-2008 : written by Pierre Chanial
;   23-Feb-2009 : PC, make the code more robust against HCSS bugs
;   modified by H.R.: added "extension" argument
;                     replaced readfits with mrdfits
;
; COPYRIGHTS: Copyright (c) 2008-2009 Pierre Chanial
;
;-

;-------------------------------------------------------------------------------
 
function read_hcss_get_keywords, header
 compile_opt hidden
 keywords = strmid(header, 0, 8)
 delim    = strmid(header, 8, 1)
 index = where(delim eq '=', count)
 keywords = strtrim(keywords[index],2)

 index = where(keywords eq 'XTENSION' or keywords eq 'EXTNAME' or              $
               keywords eq 'COMMENT' or keywords eq 'HISTORY' or               $
               keywords eq 'DSETS___' or                                       $
               stregex(keywords, '^DSET_[0-9]+ *$', /boolean), count)
 if count gt 0 then remove, index, keywords
 
 ; HCSS bug: both 'DATE-OBS' and 'DATE_OBS'
 index1 = where(keywords eq 'DATE-OBS', count1)
 index2 = where(keywords eq 'DATE_OBS', count2)
 if count1 gt 0 and count2 gt 0 then begin
    remove, index2, keywords
 endif
 
 return, idl_validname(strtrim(keywords,2), /convert_all)
end

;-------------------------------------------------------------------------------

function read_hcss_one_hdu, file, unit, extname
 compile_opt hidden
 
 ; read the data
 header = headfits(file, exten=unit)
 keywords = read_hcss_get_keywords(header)
 
 nfield = sxpar(header, 'TFIELDS', count=count)
 if count gt 0 then begin
    ____data__ = mrdfits(file, unit, /silent)
    key = strmid(keywords, 0, 5)
    index = where(key eq 'TTYPE' or key eq 'TCLAS' or key eq 'TCOLU' or key eq 'TFORM', count)
    if count gt 0 then remove, index, keywords
 endif else if sxpar(header, 'NAXIS') ne 0 then begin
    ____data__ = mrdfits(file, unit, header, /silent)
    keywords = read_hcss_get_keywords(header)
 endif
 
 if unit ne 0 then extname = idl_validname(strupcase(strtrim(sxpar(header, 'EXTNAME'),2)), /convert_all)

 ; get the FITS parameters
 for i=0, n_elements(keywords)-1 do begin
    value = sxpar(header, keywords[i])
    if size(value, /tname) eq 'STRING' then begin
       value = strtrim(value,2)
    endif
    str = '__'+keywords[i]+' = temporary(value)'
    good = execute(str)
 endfor
 
 ; do some parameter name substitutions
 index = where(strmid(header, 0, 14) eq 'HIERARCH  key.', count)
 if count gt 0 then begin
    keys = strmid(header[index], 15)
    subexpr = stregex(header[index], "^HIERARCH  key\.([A-Z0-9_-]+)='([A-Za-z0-9]+)' *$", /subexpr, /extract)
    oldkeyword = idl_validname(strupcase(subexpr[1,*]), /convert_all)
    newkeyword = idl_validname(strupcase(subexpr[2,*]), /convert_all)
    index = where(oldkeyword ne newkeyword, nkey)
    if nkey gt 0 then begin
       oldkeyword = oldkeyword[index]
       newkeyword = newkeyword[index]
       for ikey=0, nkey-1 do begin
          
          ; HCSS bug: NAXIS can be duplicated
          if stregex(oldkeyword[ikey], '^META_[0-9]+$', /boolean) and          $
             stregex(newkeyword[ikey], '^NAXIS[0-9]*$', /boolean) then begin
             continue
          endif
          
          index = where(oldkeyword[ikey] eq keywords, count)
          if count eq 0 then begin
             message, /cont, 'Keyword '+oldkeyword[ikey]+' is not found in header.'
             message, /reset
             continue
          endif
          index2 = where(newkeyword[ikey] eq keywords, count2)
          if count2 gt 0 then begin
             message, /cont, 'Keyword substitution failed: '+               $
                     oldkeyword[ikey]+'='+newkeyword[ikey]+' already exists.'
             message, /reset
             continue
          endif
          keywords[index] = newkeyword[ikey]
          str = '__'+newkeyword[ikey]+' = temporary(__'+oldkeyword[ikey]+')'
          good = execute(str)
       endfor
    endif
 endif
 
 ; deal with FITS comments and history
 __comment = sxpar(header, 'COMMENT', count=count)
 if count gt 0 then keywords = [keywords, 'COMMENT']
 __history = sxpar(header, 'HISTORY', count=count)
 if count gt 0 then keywords = [keywords, 'HISTORY']
 
 ; get the extensions
 next = sxpar(header, 'DSETS___', count=count)
 if count gt 0 then begin
    for iext=0, next-1 do begin
       hduunit = sxpar(header,'DSET_'+strtrim(iext,1), count=count)
       if count eq 0 then break
       value = read_hcss_one_hdu(file, hduunit, keyword)
       while 1 do begin
          index = where(keyword eq keywords, count)
          if count eq 0 then break
          if count gt 1 then begin
             message, 'Duplicate FITS keyword should not happen.'
          endif
          keywords[index] = keyword+'_'
          good = execute('__'+keyword+'_=temporary(__'+keyword+')')
       endwhile
       str = '__'+keyword+' = temporary(value)'
       good = execute(str)
       keywords = [keywords, keyword]
    endfor
 endif

 ____header__ = temporary(header)
 keywords = [keywords, '__header__']
 
 if __naxis ne 0 then begin
    keywords = [keywords, '__data__']
 endif
    
 str = 'result={'+strjoin(keywords+':__'+keywords, ',')+'}'
 good = execute(str)
 if good eq 0 then stop
 return, result
 
end
 
;-------------------------------------------------------------------------------
 
function read_hcss, file, extension
 compile_opt hidden
 if (not keyword_set(extension)) then extension = 0
 return, read_hcss_one_hdu(file, extension)
end
