function absdev_fit, x, y, coef=coef, ndev=ndev, noise_level=noise_level, sym=sym, dyn=dyn, ns=ns


;;; performs a robust linear fit by minimizing absolute deviations instead of squares,
;;; and iteratively excluding bright sources
;;; takes care of very large or very small dynamic ranges,
;;; but NOT intended to be universally applicable

;;; IMPORTANT: assumes that x values are sorted in ascending order

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2011 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



if not keyword_set(ndev) then ndev = 3.
if not keyword_set(ns) then ns = 25

w = where(finite(y) gt 0, complement=wz, ncomplement=cz)
min_x = min(x(w))
tab_x = x - min_x
dx = max(tab_x(w))
tab_y = y
dim = n_elements(x)
tab_dim = 1. + fltarr(dim)



;;; reduce the dynamic range to ease the fit:

if keyword_set(dyn) then begin
   tab_x_nan = tab_x
   if cz ge 1 then tab_x_nan(wz) = !values.f_nan
   m = min(abs(tab_x_nan(0:dim/3) - dx / 3.), ind_inf, /nan)
   ind = dim - 1 - dim / 3
   m = min(abs(tab_x_nan(ind:*) - 2. * dx / 3.), ind_sup, /nan)
   ind_sup += ind
   if (ind_inf ge 5 and ind_sup le dim - 6) then begin
      slope_test = (median(y(ind_sup:*), /even) - median(y(0:ind_inf), /even)) $
         / (median(tab_x_nan(ind_sup:*), /even) - median(tab_x_nan(0:ind_inf), /even))
      if finite(slope_test) gt 0 then tab_y = y - slope_test * tab_x else slope_test = 0
   endif
endif


;;; apply a first cut to bright sources:

if (not keyword_set(sym)) then begin
   histo = histogram(tab_y, locations=sb, binsize=3.*noise_level, /nan)
   histo(0) = 0.
   w = where(histo ge max(histo) / 10., cw)
   range = sb(w(cw-1)) - sb(w(0))
   thres = max([ndev, 100.]) * noise_level
   if range gt thres then begin
      wz = where(tab_y gt sb(w(0)) + thres, cz)
      if cz ge 1 then tab_y(wz) = !values.f_nan
   endif
endif


;;; to avoid infinitesimal slopes and crazy fits:

histo = histogram(tab_y, locations=sb, binsize=noise_level/3., /nan)
w = where(histo gt max(histo) / 20., cw)
dy = sb(w(cw-1)) - sb(w(0))
if dy le 0. then dy = 6. * noise_level
if (dy lt dx / 10. and (not keyword_set(dyn))) then begin
   fact_y = dx / dy
   tab_y *= fact_y
   dy *= fact_y
endif else fact_y = 1.



;;; fit with exclusion of bright sources:

tab_ns = findgen(ns) / float(ns - 1)
guess_slope = 1.2 * dy / dx
slope = -guess_slope + 2. * guess_slope * tab_ns

y_rebin = rebin(tab_y, dim, ns)
sx_rebin = (slope ## tab_dim) * rebin(tab_x, dim, ns)
norm_inv = 1. / sqrt(1. + slope^2.)


repeat begin
   cfit = median(y_rebin - sx_rebin, dimension=1, /even)
   dfit = total(abs(cfit ## tab_dim + sx_rebin - y_rebin), 1, /nan) * norm_inv

   dmin = min(dfit, ind)
   slope_fit = slope(ind)
   const = cfit(ind)
   y_fit = const + slope_fit * tab_x
   resi = tab_y - y_fit

   winc = where(finite(resi) gt 0)
   for i = 0, 6 do begin
      med = median(resi(winc))
      dev = sqrt(mean((resi(winc) - med)^2.))
      winc = where(abs(resi - med) le 3. * dev)
   endfor
   if keyword_set(sym) then w = where(abs(resi - med) gt ndev * dev, cw) $
      else w = where(resi - med gt ndev * dev, cw)
   if cw ge 1 then begin
      tab_y(w) = !values.f_nan
      y_rebin(w, *) = !values.f_nan
   endif
endrep until cw lt 1


if keyword_set(slope_test) then begin
   coef = [(const - slope_fit * min_x) / fact_y - slope_test * min_x, slope_fit / fact_y + slope_test]
   return, y_fit / fact_y + slope_test * tab_x
endif else begin
   coef = [const - slope_fit * min_x, slope_fit] / fact_y
   return, y_fit / fact_y
endelse


end
